# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
################################################################################
"""Catches vulnerable yaml desrializations that can potentially lead to
arbitrary code execution."""
from pysecsan import sanlib

try:
  import yaml
except:
  pass


def hook_pre_exec_pyyaml_load(stream, loader):
  """Hook for pyyaml.load_yaml

    Exits if the loader is unsafe or vanilla loader and the stream passed
    to the loader is controlled by the fuzz data
    """
  # Ensure loader is the unsafe loader or vanilla loader
  if (loader != yaml.loader.Loader and loader != yaml.loader.UnsafeLoader):
    return

  # Check for exact taint in stream
  if sanlib.is_exact_taint(stream):
    msg = (
        "Yaml deserialization issue.\n"
        "Unsafe deserialization can be used to execute arbitrary commands.\n")
    sanlib.abort_with_issue(msg)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
################################################################################
"""Sanitizer for regular expression dos."""

import time
import sys
from pysecsan import sanlib


# Hooks for regular expressions.
# Main problem is to identify ReDOS attemps. This is a non-trivial task
# - https://arxiv.org/pdf/1701.04045.pdf
# - https://dl.acm.org/doi/pdf/10.1145/3236024.3236027
# and the current approach we use is simply check for extensive computing time.
# In essence, this is more of a refinement of traditional timeout checker from
# the fuzzer, which, effectivelly will detect these types of attacks by way of
# timeouts.
#
# Perhaps the smartest would be to use something like e.g.
# https://github.com/doyensec/regexploit to scan the regex patterns.
# Other heuristics without going too technical on identifying super-linear
# regexes:
# - check
#   - if "taint" exists in re.compile(xx)
# - check
#   - for backtracking possbility in PATTERN within re.comile(PATTERN)
#   - and
#   - "taint" in findall(XX) calls.
def hook_post_exec_re_pattern_findall(self, s):
  global starttime
  try:
    endtime = time.time() - starttime
    if endtime > 4:
      #print("param: %s"%(s))
      raise Exception("Potential ReDOS attack")
  except NameError:
    #print("For some reason starttime is not set, which it should have")
    sys.exit(1)
    pass


def hook_pre_exec_re_pattern_findall(self, s):
  global starttime
  starttime = time.time()


def hook_post_exec_re_compile(retval, pattern, flags=None):
  """Hook for re.compile post execution to hook returned objects functions"""
  sanlib.sanitizer_log("Inside of post compile hook", 0)
  wrapper_object = sanlib.create_object_wrapper(
      findall=(hook_pre_exec_re_pattern_findall,
               hook_pre_exec_re_pattern_findall))
  hooked_object = wrapper_object(retval)
  return hooked_object


def hook_pre_exec_re_compile(pattern, flags=None):
  """Check if tainted input exists in pattern. If so, likely chance of making
    ReDOS possible."""
  sanlib.sanitizer_log("Inside re compile hook", 0)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
################################################################################
"""Sanitizers for capturing code injections"""

import sys
from typing import Optional


def get_all_substr_prefixes(main_str, sub_str):
  idx = 0
  while True:
    idx = main_str.find(sub_str, idx)
    if idx == -1:
      return
    yield main_str[0:idx]
    # Increase idx the length of the substring from the current position
    # where an occurence of the substring was found.
    idx += len(sub_str)


def check_code_injection_match(elem, check_unquoted=False) -> Optional[str]:
  # Check exact match
  if elem == "exec-sanitizer":
    return "Explicit command injection found."

  # Check potential for injecting into a string
  if "FROMFUZZ" in elem:
    if check_unquoted:
      # return true if any index is unquoted
      for sub_str in get_all_substr_prefixes(elem, "FROMFUZZ"):
        if sub_str.count("\"") % 2 == 0:
          return "Fuzzer controlled content in data. Code injection potential."

      # Return None if all fuzzer taints were quoted
      return None
    return "Fuzzer controlled content in data. Code injection potential."
  return None


def hook_pre_exec_subprocess_Popen(cmd, **kwargs):
  """Hook for subprocess.Popen"""
  if "shell" in kwargs and kwargs["shell"] is True:
    arg_shell = True
  else:
    arg_shell = False

  # Command injections depend on whether the first argument is a list of
  # strings or a string. Handle this now.
  # Example: tests/poe/ansible-runner-cve-2021-4041
  if type(cmd) is str:
    res = check_code_injection_match(cmd, check_unquoted=True)
    if res != None:
      # if shell arg is true and string is tainted and unquoted that's a
      # definite code injection.
      if arg_shell is True:
        raise Exception("Code injection in Popen")

      # Otherwise it's a maybe.
      raise Exception(f"Potental code injection in subprocess.Popen\n{res}")

  # Check for hg command injection
  # Example: tests/poe/libvcs-cve-2022-21187
  if cmd[0] == "hg":
    # Check if the arguments are controlled by the fuzzer, and this given
    # arg is not preceded by --
    found_dashes = False
    for idx in range(1, len(cmd)):
      if cmd[0] == "--":
        found_dashes = True
      if not found_dashes and check_code_injection_match(cmd[idx]):
        raise Exception(
            f"""command injection likely by way of mercurial. The following
                      command {str(cmd)} is executed, and if you substitute {cmd[idx]}
                      with \"--config=alias.init=!touch HELLO_PY\" then you will
                      create HELLO_PY""")


def hook_pre_exec_os_system(cmd):
  """Hook for os.system"""
  res = check_code_injection_match(cmd)
  if res != None:
    print("code injection by way of os.system")
    # Exceptions are not enough to throw if they are all caught: https://github.com/Lightning-AI/lightning/blob/8b7a12c52e52a06408e9231647839ddb4665e8ae/pytorch_lightning/utilities/argparse.py#L123-L125
    sys.exit(1337)
    #raise Exception(f"Potential code injection by way of os.system\n{res}")


def hook_pre_exec_eval(cmd):
  """Hook for eval. Experimental atm."""
  res = check_code_injection_match(cmd)
  if res != None:
    raise Exception(f"Potential code injection by way of eval\n{res}")



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
################################################################################

import re
import os
import sys
import time
import functools
import subprocess
import traceback
import importlib

from typing import Any, Callable, Optional
from pysecsan import command_injection, redos, yaml_deserialization

sanitizer_log_level = 0


def sanitizer_log(msg, log_level):
  global sanitizer_log_level
  if log_level >= sanitizer_log_level:
    print(f"[PYSAN] {msg}")


def is_module_present(mod_name):
  return importlib.find_loader(mod_name) is not None


def abort_with_issue(msg):
  sanitizer_log("Found an issue, pysecsan exiting", 0)
  sanitizer_log(msg, 0)
  traceback.print_stack()

  # Use os._exit here to force exit. sys.exit will exit
  # by throwing a SystemExit exception which the interpreter
  # handles by exiting. However, code may catch this exception,
  # and thus to avoid this we exit the process without exceptions.
  os._exit(1)


def is_exact_taint(stream) -> bool:
  """Checks if stream is an exact match for taint from fuzzer"""
  # The fuzzer has to get 8 characters right. This may be a bit much,
  # however, when found it shows a high level of control over the data.
  if stream == "FROMFUZZ":
    return True

  return False


def create_object_wrapper(**methods):
  """Hooks functions in an object

  This is needed for hooking built-in types and object attributes.

  Example use case is if we want to find ReDOS vulnerabilities, that
  have a pattern of

  ```
  import re
  r = re.compile(REGEX)
  for _ in r.findall(...)
  ```

  In the above case r.findall is a reference to
  re.Pattern.findall, which is a built-in type that is non-writeable.

  In order to hook such calls we need to wrap the object, and also hook the
  re.compile function to return the wrapped/hooked object.
  """

  class Wrapper(object):

    def __init__(self, instance):
      object.__setattr__(self, 'instance', instance)

    def __setattr__(self, name, value):
      object.__setattr__(object.__getattribute__(self, 'instance'), name, value)

    def __getattribute__(self, name):
      instance = object.__getattribute__(self, 'instance')

      def _hook_func(self, pre_hook, post_hook, orig, *args, **kargs):
        if pre_hook is not None:
          pre_hook(self, *args, **kargs)
        # No need to pass instance here because when we extracted
        # the funcion we used instance.__getattribute__(name) which
        # seems to include it. I think.
        r = orig(*args, **kargs)

        if post_hook is not None:
          post_hook(self, *args, **kargs)
        return r

      # If this is a wrapped method, return a bound method
      if name in methods:
        pre_hook = methods[name][0]
        post_hook = methods[name][1]
        orig = instance.__getattribute__(name)
        return (lambda *args, **kargs: _hook_func(self, pre_hook, post_hook,
                                                  orig, *args, **kargs))

      # Otherwise, just return attribute of instance
      return instance.__getattribute__(name)

  return Wrapper


def add_hook(function: Callable[[Any], Any],
             pre_exec_hook: Optional[Callable[[Any], Any]] = None,
             post_exec_hook: Optional[Callable[[Any], Any]] = None):
  """Hook a function.

    Hooks can be placed pre and post function call. At least one hook is
    needed.

    This hooking is intended on non-object hooks. In order to hook functions
    in objects the `create_object_wrapper` function is used in combination with function
    hooking initialisation functions post execution.
    """
  if pre_exec_hook is None and post_exec_hook is None:
    raise Exception("Some hooks must be included")

  @functools.wraps(function)
  def run(*args, **kwargs):
    sanitizer_log(f"Hook start {str(function)}", 0)

    # Call hook
    if pre_exec_hook is not None:
      pre_exec_hook(*args, **kwargs)

    # Call the original function in the even the hook did not indicate
    # failure.
    ret = function(*args, **kwargs)

    # Post execution hook. Overwrite return value if anything is returned
    # by post hook.
    if post_exec_hook is not None:
      tmp_ret = post_exec_hook(ret, *args, **kwargs)
      if tmp_ret is not None:
        #print("Overwriting ret value")
        ret = tmp_ret
    sanitizer_log(f"Hook end {str(function)}", 0)
    return ret

  return run


def add_hooks():
  """Sets up hooks"""
  os.system = add_hook(os.system,
                       pre_exec_hook=command_injection.hook_pre_exec_os_system)
  subprocess.Popen = add_hook(
      subprocess.Popen,
      pre_exec_hook=command_injection.hook_pre_exec_subprocess_Popen)
  re.compile = add_hook(re.compile,
                        pre_exec_hook=redos.hook_pre_exec_re_compile,
                        post_exec_hook=redos.hook_post_exec_re_compile)

  # Hack to determine if yaml is elligible, because pkg_resources does
  # not seem to work from pyinstaller.
  if is_module_present("yaml"):
    import yaml
    sanitizer_log("Hooking pyyaml.load", 0)
    yaml.load = add_hook(
        yaml.load,
        pre_exec_hook=yaml_deserialization.hook_pre_exec_pyyaml_load,
    )
  else:
    sanitizer_log("pyyaml not found. No hooks here", 0)
